const gRDFDataSource = "rdf:local-store";
const gRDFRoot = "http://click2tab.mozdev.org/rdf/all";
const gRDFNodeClosedWindows = gRDFRoot + "/window";
const gRDFLabelClosedTabs = gRDFNodeClosedWindows + "#closedtabs";
const gRDFNodeClosedTabs = gRDFNodeClosedWindows + "/closedtabs";
const gRDFNodeSession = gRDFNodeClosedTabs + "/session";
const gRDFLabelSessionIndex = gRDFNodeSession + "#index";
const gRDFLabelSessionHistory = gRDFNodeSession + "#history";
const gRDFLabelHistoryEntry = gRDFNodeSession + "/history/entry";
const gRDFLabelHistoryEntryTitle = gRDFLabelHistoryEntry + "#title";
const gRDFLabelHistoryEntryURL = gRDFLabelHistoryEntry + "#url";
const gRDFLabelHistoryEntryScrollX = gRDFLabelHistoryEntry + "#scrollx";
const gRDFLabelHistoryEntryScrollY = gRDFLabelHistoryEntry + "#scrolly";

function blankTabClick2Tab(browser)
{
    if (browser.sessionHistory.count == 0)
        return true;
    if (browser.sessionHistory.count == 1 && (browser.currentURI.spec == "about:blank"))
    {
        browser.sessionHistory.PurgeHistory(browser.sessionHistory.count);
        return true;
    }
    return false;
}

function deleteNodeClick2Tab(rdfNode, dataSource)
{
    if (!dataSource)
    {
        var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                                   .getService(Components.interfaces.nsIRDFService);
        dataSource = rdfService.GetDataSource(gRDFDataSource);
    }

    var arcOut = dataSource.ArcLabelsOut(rdfNode);
    while (arcOut.hasMoreElements())
    {
        var aLabel = arcOut.getNext();
        if (aLabel instanceof Components.interfaces.nsIRDFResource)
        {
            var aTarget = dataSource.GetTarget(rdfNode, aLabel, true);
            dataSource.Unassert(rdfNode, aLabel, aTarget);
        }
    }
}

function deleteSubtreeClick2Tab(labelRoot, dataSource)
{
    if (!dataSource)
    {
        var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                                   .getService(Components.interfaces.nsIRDFService);
        dataSource = rdfService.GetDataSource(gRDFDataSource);
    }

    var allElements = dataSource.GetAllResources();

    while (allElements.hasMoreElements())
    {
        var aResource = allElements.getNext();
        if ((aResource instanceof Components.interfaces.nsIRDFResource)
            && (aResource.Value.indexOf(labelRoot) == 0))
            deleteNodeClick2Tab(aResource, dataSource);
    }
}

var ClosedTabsClick2Tab =
{
    closedTabs : [],
    cmdUndoLastClosedTab : null,
    popupsRelated : [],
    inUpdatingPopup : false,

    setBrowserUIState : function ()
    {
        if (this.closedTabs.length)
            this.cmdUndoLastClosedTab.removeAttribute("disabled");
        else
            this.cmdUndoLastClosedTab.setAttribute("disabled", "true");

        // if this.inUpdatingPopup is true, the last request for updating popup menu has been
        // called yet, there is no need to time out a new one
        if (this.inUpdatingPopup)
            return;
        this.inUpdatingPopup = true;
        setTimeout(function () { ClosedTabsClick2Tab.updatePopupMenu(); }, 0);
    },

    updatePopupMenu : function ()
    {
        for (var i = 0; i < this.popupsRelated.length; i++)
            TabbrowserClick2Tab.buildClosedTabList(this.popupsRelated[i]);
        this.inUpdatingPopup = false;
    },

    // cache the tab if its not blank, return value indicate whether the tab is cached
    cacheClosedTab : function (browser, notUpdateUI)
    {
        if (blankTabClick2Tab(browser))
            return false;
        else
        {
            this.closedTabs.unshift([browser.sessionHistory]);
            if (this.closedTabs.length > gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedCacheSize))
                this.closedTabs.pop();

            if (!notUpdateUI)
                this.setBrowserUIState();

            return true;
        }
    },

    removeClosedTab : function (index)
    {
        this.closedTabs.splice(index, 1);
        this.setBrowserUIState();
    },

    clearAllClosedTabs : function (notUpdateUI)
    {
        this.closedTabs.splice(0, this.closedTabs.length);
        if (!notUpdateUI)
            this.setBrowserUIState();
    },

    saveClosedTabs : function ()
    {
        if (!this.closedTabs.length)
            return;

        var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                                   .getService(Components.interfaces.nsIRDFService);
        var dataSource = rdfService.GetDataSource(gRDFDataSource);
        var rdfNodeClosedWindows = rdfService.GetResource(gRDFNodeClosedWindows);
        var containerUtils = Components.classes["@mozilla.org/rdf/container-utils;1"]
                                       .getService(Components.interfaces.nsIRDFContainerUtils);

        var windowContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                        .createInstance(Components.interfaces.nsIRDFContainer);
        try
        {
            windowContainer.Init(dataSource, rdfNodeClosedWindows);
        }
        catch (ex)
        {
            windowContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedWindows);
        }

        var windowID = windowContainer.GetCount();
        var rdfNodeThisWindow = rdfService.GetResource(gRDFNodeClosedWindows + windowID);
        // Add the node to the Sequence as early as possible.
        // Not that the first element is 1 and not 0
        windowContainer.InsertElementAt(rdfNodeThisWindow, 1, true);

        var rdfLabelClosedTabs = gRDFNodeClosedWindows + windowID + "/closedtabs";
        var rdfNodeClosedTabs = rdfService.GetResource(rdfLabelClosedTabs);
        var tabContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                     .createInstance(Components.interfaces.nsIRDFContainer);
        try
        {
            tabContainer.Init(dataSource, rdfNodeClosedTabs);
        }
        catch (ex)
        {
            tabContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedTabs);
        }

        for (var i = 0; i < this.closedTabs.length; i++)
        {
            var sessionHistory = this.closedTabs[i][0];

            var rdfLabelSession = rdfLabelClosedTabs + "/session" + i;
            var rdfNodeSession = rdfService.GetResource(rdfLabelSession);

            var rdfLabelSessionHistory = rdfLabelSession + "/history";
            var rdfNodeSessionHistory = rdfService.GetResource(rdfLabelSessionHistory);
            var entryContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                           .createInstance(Components.interfaces.nsIRDFContainer);
            try
            {
                entryContainer.Init(dataSource, rdfNodeSessionHistory);
            }
            catch (ex)
            {
                entryContainer = containerUtils.MakeSeq(dataSource, rdfNodeSessionHistory);
            }

            for (var j = 0; j < sessionHistory.count; j++)
            {
                var rdfNodeHistoryEntry = rdfService.GetResource(rdfLabelSessionHistory + "/entry" + j);
                var historyEntry = sessionHistory.getEntryAtIndex(j, false);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryTitle),
                                  rdfService.GetLiteral(historyEntry.title),
                                  true);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryURL),
                                  rdfService.GetLiteral(historyEntry.URI.spec),
                                  true);
                var scrollX = {}, scrollY = {};
                historyEntry.QueryInterface(Components.interfaces.nsISHEntry).getScrollPosition(scrollX, scrollY);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryScrollX),
                                  rdfService.GetIntLiteral(scrollX.value),
                                  true);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryScrollY),
                                  rdfService.GetIntLiteral(scrollY.value),
                                  true);                

                entryContainer.AppendElement(rdfNodeHistoryEntry);
            } // end of "for (var j = 0; j < sessionHistory.count; j++)"

            dataSource.Assert(rdfNodeSession,
                              rdfService.GetResource(gRDFLabelSessionHistory),
                              rdfNodeSessionHistory,
                              true);
            dataSource.Assert(rdfNodeSession,
                              rdfService.GetResource(gRDFLabelSessionIndex),
                              rdfService.GetIntLiteral(sessionHistory.index),
                              true);

            tabContainer.AppendElement(rdfNodeSession);
        } // end of "for (var i = 0; i < this.closedTabs.length; i++)"
        dataSource.Assert(rdfNodeThisWindow,
                          rdfService.GetResource(gRDFLabelClosedTabs),
                          rdfNodeClosedTabs,
                          true);

        // Since it's possible that there are multiple windows open at the same time,
        // it would be better to save the RDF file immediately
        dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    }, // end of "saveClosedTabs : function ()"

    loadClosedTabs : function ()
    {
        var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                                   .getService(Components.interfaces.nsIRDFService);
        var dataSource = rdfService.GetDataSource(gRDFDataSource);
        var rdfNodeClosedWindows = rdfService.GetResource(gRDFNodeClosedWindows);
        var containerUtils = Components.classes["@mozilla.org/rdf/container-utils;1"]
                                       .getService(Components.interfaces.nsIRDFContainerUtils);

        var windowContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                        .createInstance(Components.interfaces.nsIRDFContainer);
        try
        {
            windowContainer.Init(dataSource, rdfNodeClosedWindows);
        }
        catch (ex)
        {
            windowContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedWindows);
        }

        var windowEnum = windowContainer.GetElements();
        if (windowEnum.hasMoreElements())
        {
            var rdfNodeLastWindow = windowEnum.getNext();
            if (rdfNodeLastWindow instanceof Components.interfaces.nsIRDFResource)
            {
                var rdfLabelClosedTabs = rdfService.GetResource(gRDFLabelClosedTabs);
                var rdfNodeClosedTabs = dataSource.GetTarget(rdfNodeLastWindow, rdfLabelClosedTabs, true);
                if (rdfNodeClosedTabs instanceof Components.interfaces.nsIRDFResource)
                {
                    var tabContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                                 .createInstance(Components.interfaces.nsIRDFContainer);
                    try
                    {
                        tabContainer.Init(dataSource, rdfNodeClosedTabs);
                    }
                    catch (ex)
                    {
                        tabContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedTabs);
                    }

                    var sessionEnum = tabContainer.GetElements();
                    while (sessionEnum.hasMoreElements())
                    {
                        var rdfNodeSession = sessionEnum.getNext();
                        if (rdfNodeSession instanceof Components.interfaces.nsIRDFResource)
                            this.loadOneTab(rdfService, dataSource, containerUtils, rdfNodeSession);
                    }
                } // end of "if (rdfNodeClosedTabs instanceof Components.interfaces.nsIRDFResource)"
                deleteSubtreeClick2Tab(rdfNodeLastWindow.Value, dataSource);
                windowContainer.RemoveElement(rdfNodeLastWindow, true);
                // delete the sequence if blank
                if (!windowContainer.GetCount())
                    deleteNodeClick2Tab(rdfNodeClosedWindows, dataSource);

                // it would be better to save the RDF file immediately
                dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();

                this.setBrowserUIState();
            } // end of "if (rdfNodeLastWindow instanceof Components.interfaces.nsIRDFResource)"
        } // end of "if (windowEnum.hasMoreElements())"
    }, // end of "loadClosedTabs : function ()"

    loadOneTab : function(rdfService, dataSource, containerUtils, rdfNodeSession)
    {
        var rdfLabelSessionHistory = rdfService.GetResource(gRDFLabelSessionHistory);
        var rdfNodeSessionHistory = dataSource.GetTarget(rdfNodeSession, rdfLabelSessionHistory, true);
        if (rdfNodeSessionHistory instanceof Components.interfaces.nsIRDFResource)
        {
            var entryContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                           .createInstance(Components.interfaces.nsIRDFContainer);
            try
            {
                entryContainer.Init(dataSource, rdfNodeSessionHistory);
            }
            catch (ex)
            {
                entryContainer = containerUtils.MakeSeq(dataSource, rdfNodeSessionHistory);
            }

            var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                                      .getService(Components.interfaces.nsIIOService);

            var sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"]
                                           .createInstance(Components.interfaces.nsISHistory);
            var sHistoryInternal = sessionHistory;
            sHistoryInternal = sHistoryInternal.QueryInterface(Components.interfaces.nsISHistoryInternal);

            var entryEnum = entryContainer.GetElements();
            while (entryEnum.hasMoreElements())
            {
                var rdfNodeHistoryEntry = entryEnum.getNext();
                if (rdfNodeHistoryEntry instanceof Components.interfaces.nsIRDFResource)
                {
                    var historyEntry = Components.classes["@mozilla.org/browser/session-history-entry;1"]
                                                 .createInstance(Components.interfaces.nsISHEntry);

                    var rdfLabelHistoryEntryTitle = rdfService.GetResource(gRDFLabelHistoryEntryTitle);
                    var rdfNodeHistoryEntryTitle = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryTitle, true);
                    var entryTitle = "";
                    if (rdfNodeHistoryEntryTitle instanceof Components.interfaces.nsIRDFLiteral)
                        entryTitle = rdfNodeHistoryEntryTitle.Value;
                    // Method "setTitle" name changed to "SetTitle" since Trunk build 20050515, 
                    // the following code block is used to make it compatible with both branch and trunk build 
                    if ("SetTitle" in historyEntry)
                        historyEntry.SetTitle(entryTitle);
                    else
                        historyEntry.setTitle(entryTitle);

                    var rdfLabelHistoryEntryURL = rdfService.GetResource(gRDFLabelHistoryEntryURL);
                    var rdfNodeHistoryEntryURL = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryURL, true);
                    var entryURL = "about:blank";
                    if (rdfNodeHistoryEntryURL instanceof Components.interfaces.nsIRDFLiteral)
                        entryURL = rdfNodeHistoryEntryURL.Value;
                    if ("SetURI" in historyEntry)
                        historyEntry.SetURI(ioService.newURI(entryURL, null, null));
                    else
                        historyEntry.setURI(ioService.newURI(entryURL, null, null));

                    var rdfLabelHistoryEntryScrollX = rdfService.GetResource(gRDFLabelHistoryEntryScrollX);
                    var rdfNodeHistoryEntryScrollX = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryScrollX, true);
                    var entryScrollX = 0;
                    if (rdfNodeHistoryEntryScrollX instanceof Components.interfaces.nsIRDFInt)
                        entryScrollX = rdfNodeHistoryEntryScrollX.Value;

                    var rdfLabelHistoryEntryScrollY = rdfService.GetResource(gRDFLabelHistoryEntryScrollY);
                    var rdfNodeHistoryEntryScrollY = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryScrollY, true);
                    var entryScrollY = 0;
                    if (rdfNodeHistoryEntryScrollY instanceof Components.interfaces.nsIRDFInt)
                        entryScrollY = rdfNodeHistoryEntryScrollY.Value;

                    historyEntry.setScrollPosition(entryScrollX, entryScrollY);

                    sHistoryInternal.addEntry(historyEntry, true);
                } // end of "if (rdfNodeHistoryEntry instanceof Components.interfaces.nsIRDFResource)"
            } // end of "while (entryEnum.hasMoreElements())"

            var rdfLabelSessionIndex = rdfService.GetResource(gRDFLabelSessionIndex);
            var rdfNodeSessionIndex = dataSource.GetTarget(rdfNodeSession, rdfLabelSessionIndex, true);
            var sessionIndex = sessionHistory.count - 1;
            if (rdfNodeSessionIndex instanceof Components.interfaces.nsIRDFInt)
                sessionIndex = rdfNodeSessionIndex.Value;
            // Set the current history index with the second parameter set to true
            sessionHistory.getEntryAtIndex(sessionIndex, true);

            this.closedTabs.push([sessionHistory]);
        } // end of "if (rdfNodeSessionHistory instanceof Components.interfaces.nsIRDFResource)"
    } // end of "loadOneTab : function(rdfService, rdfDatasource, rdfNodeSession)"
};

const Click2TabSettingsListener = 
{
    observe : function (subject, topic, prefName)
    {
        if (topic != "nsPref:changed")
            return;

        switch (prefName)
        {
            case prefStringClick2TabCloseTabButtonContext:
                TabbrowserClick2Tab.onCloseTabButtonContextChanged();
                break;
            case prefStringClick2TabReopenTabContext:
                TabbrowserClick2Tab.onReopenTabContextChanged();
                break;
            case prefStringClick2TabReopenTabsTabContext:
                TabbrowserClick2Tab.onReopenTabsTabContextChanged();
                break;
            case prefStringClick2TabCloseTabButtonMiddleClick:
                TabbrowserClick2Tab.onCloseTabButtonMiddleClickChanged();
                break;
            case prefStringClick2TabClosedCacheSize:
                TabbrowserClick2Tab.onClosedCacheSizeChanged();
                break;
            case prefStringClick2TabClosedListLength:
                TabbrowserClick2Tab.onClosedListLengthChanged();
                break;
            case prefStringClick2TabIndexInTabList:
                TabbrowserClick2Tab.onTabIndexInTabListChanged();
            case prefStringClick2TabShowMoreSubmenu:
                TabbrowserClick2Tab.onShowMoreSubmenuChanged();
            default:
                break;
        }
    }
};

var TabbrowserClick2Tab =
{
    prefBranch : null,

    closeTabButton : null,

    closeTabButtonContext : null,
    closeTabButtonContextReplaced : false,
    closeTabButtonOnClick : null,
    closeTabButtonOnClickReplaced : false,

    menupopupTabContext : null,
    tabContextMenuId : null,

    reopenMenuSep : null,
    reopenLastTabMenuItem : null,
    reopenTabsMenu : null,

    // append the three menu items when the tab context menu is first shown
    reopenMenusAppended : false,

    stripTooltipOnPopupShowing : null,

    // Since it's possbile for the UNDO menu-button to be invisible, we can't regisiter its
    // popup on initialization. But we can register the popup on the first call to its
    // "onpopupshowing" event handler. This boolean value is used to indicate whether the
    // popup has been registered to avoid multiple registration of it.
    undoButtonPopupRegistered : false,

    // Get the id of the tab bar and tab button context menu.
    // If the menu has no id, set one for it
    getTabContextMenuId : function ()
    {
        if (this.menupopupTabContext.hasAttribute("id"))
            this.tabContextMenuId = this.menupopupTabContext.getAttribute("id");
        else
        {
            this.tabContextMenuId = "tab-context-menu-click2tab";
            this.menupopupTabContext.setAttribute("id", this.tabContextMenuId);
        }
    },

    getCloseTabButtonContext : function ()
    {
        // In recent trunk builds, the Close Tab Button has been moved from the right side of the tab bar to each each tab
        // button, so the current method for replacing the context menu of Close Tab Button doesn't work anymore (the context
        // menu for each Close Tab Button should be replaced. This change also fails the initialization of Click2Tab (on startup
        // there is only one tab and the Close Tab Button will not be created).
        // 
        // The method for replacing the context menu (and middle button click event handler, for the same reason) should be
        // improvement to make it work in trunk build. Currently I would only make the other feature functional again.
        //
        // 2006/02/11
        if (this.closeTabButton == null)
            return;

        this.getTabContextMenuId();
        if (this.closeTabButton.hasAttribute("context"))
            this.closeTabButtonContext = this.closeTabButton.getAttribute("context");
        else
            this.closeTabButtonContext = this.tabContextMenuId;
    },

    replaceCloseTabButtonContext : function ()
    {
        if (this.closeTabButton == null)
            return;

        this.closeTabButton.setAttribute("context", "undo-closetabs-popup");
        this.closeTabButtonContextReplaced = true;
    },

    restoreCloseTabButtonContext : function ()
    {
        if (this.closeTabButton == null)
            return;

        if (this.closeTabButtonContextReplaced)
        {
            this.closeTabButton.setAttribute("context", this.closeTabButtonContext);
            this.closeTabButtonContextReplaced = false;
        }
    },

    getcloseTabButtonOnClick : function ()
    {
        if (this.closeTabButton == null)
            return;

        if (this.closeTabButton.hasAttribute("onclick"))
            this.closeTabButtonOnClick = this.closeTabButton.getAttribute("onclick");
        else
            this.closeTabButtonOnClick = "return;";
    },

    replaceCloseTabButtonOnClick : function ()
    {
        if (this.closeTabButton == null)
            return;

        this.closeTabButton.setAttribute("onclick", "TabbrowserClick2Tab.onCloseTabButtonClick(event);");
        this.closeTabButtonOnClickReplaced = true;
    },

    restoreCloseTabButtonOnClick : function ()
    {
        if (this.closeTabButton == null)
            return;

        if (this.closeTabButtonOnClickReplaced)
        {
            this.closeTabButton.setAttribute("onclick", this.closeTabButtonOnClick);
            this.closeTabButtonOnClickReplaced = false;
        }
    },

    addReopenToTabContext : function ()
    {
        this.reopenMenuSep = document.createElement("menuseparator");
        this.reopenMenuSep.setAttribute("hidden", "true");

        this.reopenLastTabMenuItem = document.getElementById("reopen-lastclosedtab-menuitem");
        this.reopenTabsMenu = document.getElementById("undo-closedtabs-menu");

        this.menupopupTabContext.addEventListener("popupshowing", function () { TabbrowserClick2Tab.appendReopenMenus(); }, false);
    },

    appendReopenMenus : function ()
    {
        if (!this.reopenMenusAppended)
        {
            this.reopenMenusAppended = true;
            //this.menupopupTabContext.removeEventListener("popupshowing", function () { TabbrowserClick2Tab.appendReopenMenus(); }, false);
            this.menupopupTabContext.appendChild(this.reopenMenuSep);
            this.menupopupTabContext.appendChild(this.reopenLastTabMenuItem);
            this.menupopupTabContext.appendChild(this.reopenTabsMenu);
        }
    },

    showReopenInTabContext : function ()
    {
        this.reopenMenuSep.removeAttribute("hidden");
        this.reopenLastTabMenuItem.removeAttribute("hidden");
    },

    hideReopenInTabContext : function ()
    {
        if (this.reopenTabsMenu.hasAttribute("hidden")
            && (this.reopenTabsMenu.getAttribute("hidden") == "true"))
            this.reopenMenuSep.setAttribute("hidden", "true");
        this.reopenLastTabMenuItem.setAttribute("hidden", "true");
    },

    showReopenTabsTabContext : function ()
    {
        this.reopenMenuSep.removeAttribute("hidden");
        this.reopenTabsMenu.removeAttribute("hidden");
    },

    hideReopenTabsTabContext : function ()
    {
        if (this.reopenLastTabMenuItem.hasAttribute("hidden")
            && (this.reopenLastTabMenuItem.getAttribute("hidden") == "true"))
            this.reopenMenuSep.setAttribute("hidden", "true");
        this.reopenTabsMenu.setAttribute("hidden", "true");
    },

    removeReopenToTabContext : function ()
    {
        if (this.reopenMenusAppended)
        {
            this.menupopupTabContext.removeChild(this.reopenMenuSep);
            this.menupopupTabContext.removeChild(this.reopenLastTabMenuItem);
            this.menupopupTabContext.removeChild(this.reopenTabsMenu);
        }
        //else
        //    this.menupopupTabContext.removeEventListener("popupshowing", function () { TabbrowserClick2Tab.appendReopenMenus(); }, false);
    },

    setShortcutKeyUndo : function ()
    {
        var keyUndoLastClosedTab = document.getElementById("key_undoLastClosedTab");

        var keys = gBrowser.mPrefs.getCharPref(prefStringClick2TabShortcutKeyUndo).split("][");
        if (keys[0])
            keyUndoLastClosedTab.setAttribute("modifiers", keys[0]);
        if (keys[1])
            keyUndoLastClosedTab.setAttribute("key", keys[1]);
        if (keys[2])
            keyUndoLastClosedTab.setAttribute("keycode", keys[2]);
        if (keys[1] || keys[2])
            keyUndoLastClosedTab.setAttribute("command", "cmd_undoLastClosedTab");
    },

    removeShortcutKeyUndo : function ()
    {
        var keyUndoLastClosedTab = document.getElementById("key_undoLastClosedTab");

        if (keyUndoLastClosedTab.hasAttribute("modifiers"))
            keyUndoLastClosedTab.removeAttribute("modifiers");
        if (keyUndoLastClosedTab.hasAttribute("key"))
            keyUndoLastClosedTab.removeAttribute("key");
        if (keyUndoLastClosedTab.hasAttribute("keycode"))
            keyUndoLastClosedTab.removeAttribute("keycode");
        if (keyUndoLastClosedTab.hasAttribute("command"))
            keyUndoLastClosedTab.removeAttribute("command");
    },

    changeStripTooltip : function ()
    {
        var stripTooltip = gBrowser.mStrip.childNodes[0];

        this.stripTooltipOnPopupShowing = stripTooltip.getAttribute("onpopupshowing");
        var onStripTooltipPopupShowing = this.stripTooltipOnPopupShowing.replace(
                                            "if (document.tooltipNode.hasAttribute('label'))",
                                            "if (document.tooltipNode.hasAttribute('label') && !document.tooltipNode.hasAttribute('tooltiptext'))");
        stripTooltip.setAttribute("onpopupshowing", onStripTooltipPopupShowing);
    },

    restoreStripTooltop : function ()
    {
        var stripTooltip = gBrowser.mStrip.childNodes[0];

        stripTooltip.setAttribute("onpopupshowing", this.stripTooltipOnPopupShowing);
    },

    init : function ()
    {
        this.prefBranch = gBrowser.mPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
        this.closeTabButton = document.getAnonymousElementByAttribute(gBrowser.mTabContainer,
                                                                    "class",
                                                                    "tabs-closebutton close-button");
        this.menupopupTabContext = document.getAnonymousElementByAttribute(gBrowser,
                                                                           "anonid",
                                                                           "tabContextMenu");

        this.getCloseTabButtonContext();
        this.getcloseTabButtonOnClick();

        this.prefBranch.addObserver(prefStringClick2TabReopenTabContext, Click2TabSettingsListener, false);
        this.prefBranch.addObserver(prefStringClick2TabReopenTabsTabContext, Click2TabSettingsListener, false);
        this.addReopenToTabContext();
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabContext))
            this.showReopenInTabContext();
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabsTabContext))
            this.showReopenTabsTabContext();

        // Change the context menu of Close Tab Button. Thanks to the idea of Twanno!
        this.prefBranch.addObserver(prefStringClick2TabCloseTabButtonContext, Click2TabSettingsListener, false);
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonContext))
            this.replaceCloseTabButtonContext();

        this.prefBranch.addObserver(prefStringClick2TabCloseTabButtonMiddleClick, Click2TabSettingsListener, false);
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonMiddleClick))
            this.replaceCloseTabButtonOnClick();

        this.setShortcutKeyUndo();
        this.changeStripTooltip();

        this.prefBranch.addObserver(prefStringClick2TabClosedCacheSize, Click2TabSettingsListener, false);
        this.prefBranch.addObserver(prefStringClick2TabClosedListLength, Click2TabSettingsListener, false);
        this.prefBranch.addObserver(prefStringClick2TabIndexInTabList, Click2TabSettingsListener, false);
        this.prefBranch.addObserver(prefStringClick2TabShowMoreSubmenu, Click2TabSettingsListener, false);

        ClosedTabsClick2Tab.cmdUndoLastClosedTab = document.getElementById("cmd_undoLastClosedTab");
        ClosedTabsClick2Tab.popupsRelated.push(document.getElementById("undo-closetabs-popup"));
        ClosedTabsClick2Tab.popupsRelated.push(document.getElementById("undo-closetabs-menupopup"));

        // The loading of closed tabs should also be executed as late as possible for the UI
        // element to get ready
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabSaveClosedTabs)
            || gBrowser.mPrefs.getBoolPref(prefStringClick2TabSaveOpenTabs))
            ClosedTabsClick2Tab.loadClosedTabs();

        // It's more secure to replace those default functions in the last step of initializtion
        var regexp_in_removeTab = /this.setStripVisibilityTo\(false\);([ \r\n])*return;/;
        eval("gBrowser.removeTab = " +
              // Since Firefox Trunk build 20051015, the "toString" function adds the string "function::"
              // before every call to the member functions of JS object. The next line of code is used to
              // remove it to make the followed code replacement functional.
              gBrowser.removeTab.toString().replace(/\.function::/g, '.')
                                           .replace('if (l == 1) {',
                                                    'if (l == 1) { \
                                                         var mainWindow = document.getElementById("main-window"); \
                                                         if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCtlWCloseTab) \
                                                             && !(mainWindow.hasAttribute("chromehidden") && (mainWindow.getAttribute("chromehidden").indexOf("location") != -1))) {')
                                           .replace('if (!this.mPrefs.getBoolPref("browser.tabs.autoHide")) {',
                                                    '{ \
                                                         if (this.mPrefs.getBoolPref(prefStringClick2TabClearHistory)) \
                                                             if (ClosedTabsClick2Tab.cacheClosedTab(this.mCurrentBrowser)) \
                                                                 this.webNavigation.sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"] \
                                                                                                               .createInstance(Components.interfaces.nsISHistory);')
                                           .replace('this.loadURI("about:blank");',
                                                    'this.loadURI("about:blank"); \
                                                     if (gURLBar) \
                                                         setTimeout(function() { gURLBar.focus(); }, 0);')
                                           .replace(regexp_in_removeTab,
                                                    '   this.setStripVisibilityTo(false); \
                                                        return; \
                                                     } \
                                                     else \
                                                     { \
                                                         BrowserCloseWindow(); \
                                                         return; \
                                                     }')
                                           .replace('var oldTab = aTab;',
                                                    'ClosedTabsClick2Tab.cacheClosedTab(oldBrowser); \
                                                     var oldTab = aTab;'));

        eval('BrowserCloseTabOrWindow = ' +
             BrowserCloseTabOrWindow.toString().replace('gBrowser.tabContainer.childNodes.length > 1',
                                                        'true'));

        // the updateGoMenu function is still overlayed in this manner since it's not
        // a usually overlayed function in extensions and the code is really different from
        // the original one
        __updateGoMenu = updateGoMenu;
        updateGoMenu = this.updateGoMenu;
    },

    deinit : function ()
    {
        updateGoMenu = __updateGoMenu;
        eval("BrowserCloseTabOrWindow = " +
             BrowserCloseTabOrWindow.toString().replace('true',
                                                        'gBrowser.tabContainer.childNodes.length > 1'));
        // With the 'replace' method for overlay of 'removeTab', it's a little difficult to restore the 
        // original 'removeTab' function in a conforming way. So temporarily abandon the restoration.

        if (!gBrowser.mPrefs.getBoolPref(prefStringClick2TabSaveClosedTabs))
            ClosedTabsClick2Tab.clearAllClosedTabs(true);
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabSaveOpenTabs))
            TabbrowserClick2Tab.cacheAllOpenTabs();
        ClosedTabsClick2Tab.saveClosedTabs();

        this.prefBranch.removeObserver(prefStringClick2TabShowMoreSubmenu, Click2TabSettingsListener);
        this.prefBranch.removeObserver(prefStringClick2TabIndexInTabList, Click2TabSettingsListener);
        this.prefBranch.removeObserver(prefStringClick2TabClosedListLength, Click2TabSettingsListener);
        this.prefBranch.removeObserver(prefStringClick2TabClosedCacheSize, Click2TabSettingsListener);

        this.restoreStripTooltop();
        this.removeShortcutKeyUndo();

        this.restoreCloseTabButtonOnClick();
        this.prefBranch.removeObserver(prefStringClick2TabCloseTabButtonMiddleClick, Click2TabSettingsListener);

        this.restoreCloseTabButtonContext();
        this.prefBranch.removeObserver(prefStringClick2TabCloseTabButtonContext, Click2TabSettingsListener);

        this.removeReopenToTabContext();
        this.prefBranch.removeObserver(prefStringClick2TabReopenTabContext, Click2TabSettingsListener);
        this.prefBranch.removeObserver(prefStringClick2TabReopenTabsTabContext, Click2TabSettingsListener);
    },

    onCloseTabButtonContextChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonContext))
            this.replaceCloseTabButtonContext();
        else
            this.restoreCloseTabButtonContext ();
    },

    onReopenTabContextChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabContext))
            this.showReopenInTabContext();
        else
            this.hideReopenInTabContext();
    },

    onReopenTabsTabContextChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabsTabContext))
            this.showReopenTabsTabContext();
        else
            this.hideReopenTabsTabContext();
    },

    onCloseTabButtonMiddleClickChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonMiddleClick))
            this.replaceCloseTabButtonOnClick();
        else
            this.restoreCloseTabButtonOnClick();
    },

    onClosedCacheSizeChanged : function ()
    {
        var closedTabsCacheSize = gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedCacheSize);
        if (ClosedTabsClick2Tab.closedTabs.length > closedTabsCacheSize)
        {
            ClosedTabsClick2Tab.closedTabs.splice(closedTabsCacheSize,
                                                  ClosedTabsClick2Tab.closedTabs.length - closedTabsCacheSize);
            ClosedTabsClick2Tab.setBrowserUIState();
        }
    },

    onClosedListLengthChanged : function ()
    {
        ClosedTabsClick2Tab.setBrowserUIState();
    },

    onTabIndexInTabListChanged : function ()
    {
        ClosedTabsClick2Tab.setBrowserUIState();
    },

    onShowMoreSubmenuChanged : function ()
    {
        if (ClosedTabsClick2Tab.closedTabs.length > gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedListLength))
            ClosedTabsClick2Tab.setBrowserUIState();
    },

    onCloseTabButtonClick : function (event)
    {
        if (event.button == 1)
            this.undoLastClosedTab();
    },

    // The 'replace' method is used to make a better compatibility with other extensions. So the following function
    // (the function after the replacement) is not needed. Just keep it here for reference.
    /*removeTab : function (aTab)
    {
        if (aTab.localName != "tab")
            aTab = this.mCurrentTab;

        var l = this.mTabContainer.childNodes.length;
        if (l == 1)
        {
            var mainWindow = document.getElementById("main-window");
            // the preference "click2tab.tab.ctlWCloseTab" is now used to toggle whether the Firefox window
            // will be closed when the last tab is closed (no matter which way used to close)
            // ONE EXCEPTION: when the location bar is hidden, the window will probably a popup window. 
            //                Thus the window will be closed, regardless of the user settings.
            if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCtlWCloseTab)
                && !(mainWindow.hasAttribute("chromehidden") && (mainWindow.getAttribute("chromehidden").indexOf("location") != -1)))
            {
                //if (!this.mPrefs.getBoolPref("browser.tabs.autoHide")) 
                {
                    // Clear the session history of the last tab, added by NetOne on 2004/11/16
                    if (this.mPrefs.getBoolPref(prefStringClick2TabClearHistory))
                        // cache the session history for undo closed tabs, added by NetOne on 2004/12/12
                        if (ClosedTabsClick2Tab.cacheClosedTab(this.mCurrentBrowser))
                            // webNavigation must be used since sessionHistory is a readonly property of gBrowser
                            this.webNavigation.sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"]                        
                                                                          .createInstance(Components.interfaces.nsISHistory);

                    // blank the tab
                    this.loadURI("about:blank");
                    // Set focus to location bar, added by NetOne on 2005/02/20
                    if (gURLBar)
                        setTimeout(function() { gURLBar.focus(); }, 0);
                    return;
                }
                // hide the tab bar
                this.mPrefs.setBoolPref("browser.tabs.forceHide", true);
                this.setStripVisibilityTo(false);
                return;
            }
            else
            {
                BrowserCloseWindow();
                return;
            }
        }

        var ds = this.getBrowserForTab(aTab).docShell;

        if (ds.contentViewer && !ds.contentViewer.permitUnload())
            return;

        if (l == 2) {
            var autohide = this.mPrefs.getBoolPref("browser.tabs.autoHide");
            if (autohide)
                this.setStripVisibilityTo(false);
        }

        var index = -1;
        if (this.mCurrentTab == aTab)
            index = this.mPanelContainer.selectedIndex;
        else {
            // Find and locate the tab in our list.
            for (var i = 0; i < l; i++)
                if (this.mTabContainer.childNodes[i] == aTab)
                    index = i;
        }

        // Remove the tab's filter and progress listener.
        const filter = this.mTabFilters[index];
        var oldBrowser = this.getBrowserAtIndex(index);
        oldBrowser.webProgress.removeProgressListener(filter);
        filter.removeProgressListener(this.mTabListeners[index]);
        this.mTabFilters.splice(index, 1);
        this.mTabListeners.splice(index, 1);

        // Remove our title change and blocking listeners
        oldBrowser.removeEventListener("DOMTitleChanged", this.onTitleChanged, false);

        // We are no longer the primary content area.
        oldBrowser.setAttribute("type", "content");

        // Now select the new tab before nuking the old one.
        var currentIndex = this.mPanelContainer.selectedIndex;

        var newIndex = -1;
        if (currentIndex > index)
            newIndex = currentIndex-1;
        else if (currentIndex < index)
            newIndex = currentIndex;
        else if (index == l - 1)
            newIndex = index-1;
        else
            newIndex = index;

        // cache the session history for undo closed tabs, added by NetOne on 2004/12/12
        ClosedTabsClick2Tab.cacheClosedTab(oldBrowser);

        var oldTab = aTab;

        // clean up the before/afterselected attributes before removing the tab
        oldTab.selected = false;

        // XXX browser's destructor isn't always called, so we force a cleanup ourselves
        oldBrowser.destroy();

        this.mTabContainer.removeChild(oldTab);
        this.mPanelContainer.removeChild(this.mPanelContainer.childNodes[index]);

        this.selectedTab = this.mTabContainer.childNodes[newIndex];
        this.mPanelContainer.selectedIndex = newIndex;

        this.updateCurrentBrowser();
    },*/

    onUndoClosedTab : function (element)
    {
        if (element.hasAttribute("closedTabsIndex"))
            this.undoClosedTab(element.getAttribute("closedTabsIndex"));
    },

    undoClosedTab : function (closedTabsIndex)
    {
        if (!ClosedTabsClick2Tab.closedTabs.length)
            return;

        var oldHistory = ClosedTabsClick2Tab.closedTabs[closedTabsIndex][0];
        ClosedTabsClick2Tab.removeClosedTab(closedTabsIndex);

        var aTab, aBrowser;
        if (gBrowser.webProgress.isLoadingDocument || !blankTabClick2Tab(gBrowser.mCurrentBrowser))
        {
            aTab = gBrowser.addTab("about:blank");
            if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenInBackground))
                aBrowser = gBrowser.getBrowserForTab(aTab);
            else
            {
                gBrowser.selectedTab = aTab;
                aBrowser = gBrowser.mCurrentBrowser;
            }
        }
        else
            aBrowser = gBrowser.mCurrentBrowser;

        var newHistory = aBrowser.sessionHistory;
        newHistory.QueryInterface(Components.interfaces.nsISHistoryInternal);
        for (var i = 0; i < oldHistory.count; i++)
        {
            newHistory.addEntry(oldHistory.getEntryAtIndex(i, false), true);
        }
        if (oldHistory.count)
            aBrowser.webNavigation.gotoIndex(oldHistory.index);
    },

    undoLastClosedTab : function ()
    {
        this.undoClosedTab(0);
    },

    // This event handler doesn't need to return any value, since when it's called,
    // command "cmd_undoLastClosedTab" must have been enabled and there must be
    // at least one closed tab that needs to be displayed.
    onUndoButtonPopupShowing : function (undoButtonPopup)
    {
        if (!this.undoButtonPopupRegistered)
        {
            this.buildClosedTabList(undoButtonPopup);
            ClosedTabsClick2Tab.popupsRelated.push(undoButtonPopup);
            this.undoButtonPopupRegistered = true;
        }
    },

    updateGoMenu : function (goMenu)
    {
        if (!gBrowser.mPrefs.getBoolPref(prefStringClick2TabReplaceGoMenuItem))
            return __updateGoMenu(goMenu);

        // In case the timer didn't fire.
        destroyGoMenuItems(goMenu);

        var endSeparator = document.getElementById("endHistorySeparator");
        var closedTabsListLength = gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedListLength);
        for (var i = 0; i < closedTabsListLength && i < ClosedTabsClick2Tab.closedTabs.length; i++)
        {
            var menuItem = TabbrowserClick2Tab.constructTabMenuItem(i);
            menuItem.setAttribute("oncommand",
                                  "TabbrowserClick2Tab.onUndoClosedTab(this); event.preventBubble();");
            menuItem.setAttribute("onclick", "TabbrowserClick2Tab.onTabMenuItemClick(this, event);");
            goMenu.insertBefore(menuItem, endSeparator);

            endSeparator.hidden = false;
        }
    },

    destroyClosedTabList : function (popup)
    {
        while (popup.childNodes.length > 3)
            popup.removeChild(popup.firstChild);

        var menuMorePopup = popup.childNodes[1].firstChild;
        while (menuMorePopup.hasChildNodes())
            menuMorePopup.removeChild(menuMorePopup.firstChild);
    },

    buildClosedTabList : function (popup)
    {
        var menuItem;
        var i;

        this.destroyClosedTabList(popup);

        var menuSeparator = popup.childNodes[0],
            menuMore = popup.childNodes[1],
            menuItemClearCache = popup.childNodes[2];
        var closedTabsListLength = gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedListLength);
        for (i = 0; i < closedTabsListLength && i < ClosedTabsClick2Tab.closedTabs.length; i++)
            popup.insertBefore(this.constructTabMenuItem(i), menuSeparator);

        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabShowMoreSubmenu)
            && (i < ClosedTabsClick2Tab.closedTabs.length))
        {
            var menuMorePopup = menuMore.firstChild;
            for (; i < ClosedTabsClick2Tab.closedTabs.length; i++)
                menuMorePopup.appendChild(this.constructTabMenuItem(i));

            menuMore.removeAttribute("hidden");
        }
        else
            menuMore.setAttribute("hidden", "true");
    },

    constructTabMenuItem : function (closedTabsIndex)
    {
        var menuItem = document.createElement("menuitem");
        var tabHistory = ClosedTabsClick2Tab.closedTabs[closedTabsIndex][0];
        var title = tabHistory.getEntryAtIndex(tabHistory.index, false).title;
        var uri = tabHistory.getEntryAtIndex(tabHistory.index, false).URI.spec;

        var labelPrefix = "";
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabIndexInTabList))
        {
            if (closedTabsIndex < 10)
                menuItem.setAttribute("accesskey", closedTabsIndex);
            labelPrefix = closedTabsIndex + ". ";
        }
        menuItem.setAttribute("label", labelPrefix + (title || uri));
        menuItem.setAttribute("tooltiptext", title ? uri : "");

        menuItem.setAttribute("closedTabsIndex", closedTabsIndex);

        return menuItem;
    },

    // Closes all popups that are ancestors of the node
    // The original version can't hide popup with "xul:menupopup" and "xul:popup" as tagName
    closeMenus : function (node)
    {
        if ("tagName" in node) 
        {
            if ((node.namespaceURI == "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul")
                && ((node.tagName == "menupopup") || (node.tagName == "popup")
                    || (node.tagName == "xul:menupopup") || (node.tagName == "xul:popup")))
                node.hidePopup();

            this.closeMenus(node.parentNode);
        }
    },

    onTabMenuItemClick : function (event)
    {
        var menuItem = event.originalTarget;

        if (event.button == 2)
            if (!menuItem.hasAttribute("closedTabsIndex")
                || !gBrowser.mPrefs.getBoolPref(prefStringClick2TabRightClickTabToRemove))
                event.preventDefault();
            else
            {
                this.closeMenus(menuItem);
                // in the "else" clause, the menuItem must have the "closedTabsIndex" attribute
                ClosedTabsClick2Tab.removeClosedTab(menuItem.getAttribute("closedTabsIndex"));
            }
    },

    // The 'replace' method is used to make a better compatibility with other extensions. So the following function
    // (the function after the replacement) is not needed. Just keep it here for reference.
    /*BrowserCloseTabOrWindow : function ()
    {
        // Code changed by Netone to allow Ctrl+W only close the tab, on 12/11/2004 and 01/09/2005
        // Just close up a tab in any of the following three cases:
        // 1. There are more than one tabs opened
        // 2. For a non-popup window, there are only one tab opened and the user only want the tab being closed
        if ((gBrowser.localName == 'tabbrowser')
            && ((gBrowser.tabContainer.childNodes.length > 1)
                || (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCtlWCloseTab))))
        {
            gBrowser.removeCurrentTab();
            return;
        }

        BrowserCloseWindow();
    },*/

    cacheAllOpenTabs : function ()
    {
        for (var i = gBrowser.tabContainer.childNodes.length - 1; i >= 0; i--)
            ClosedTabsClick2Tab.cacheClosedTab(gBrowser.getBrowserAtIndex(i), true);
    }
};
